import os
import datetime
import logging
import logging.config
import yaml
import redis
import uuid
import traceback
import threading

import consul
import pika
from flask import Flask
from flask import jsonify

from biomaj_download.download.curl import CurlDownload
from biomaj_download.download.direct import DirectFTPDownload
from biomaj_download.download.direct import DirectHTTPDownload
from biomaj_download.download.localcopy import LocalDownload
from biomaj_download.message import downmessage_pb2
from biomaj_download.download.rsync import RSYNCDownload
from biomaj_core.utils import Utils
from biomaj_zipkin.zipkin import Zipkin
from biomaj_download.download.protocolirods import IRODSDownload

app = Flask(__name__)
app_log = logging.getLogger('werkzeug')
app_log.setLevel(logging.ERROR)

# Classify protocols from downmessage.proto
# Note: those lists are based on the protocol numbers, not the protocol names
ALL_PROTOCOLS = [item for key, item in downmessage_pb2.DownloadFile.Protocol.items()]
DIRECT_PROTOCOLS = [
    item for key, item in downmessage_pb2.DownloadFile.Protocol.items()
    if key.startswith("DIRECT")
]


@app.route('/api/download-message')
def ping():
    return jsonify({'msg': 'pong'})


def start_web(config):
    app.run(host='0.0.0.0', port=config['web']['port'])


def consul_declare(config):
    if config['consul']['host']:
        consul_agent = consul.Consul(host=config['consul']['host'])
        consul_agent.agent.service.register(
            'biomaj-download-message',
            service_id=config['consul']['id'],
            address=config['web']['hostname'],
            port=config['web']['port'],
            tags=['biomaj']
        )
        check = consul.Check.http(
            url='http://' + config['web']['hostname'] + ':' + str(config['web']['port']) + '/api/download-message',
            interval=20
        )
        consul_agent.agent.check.register(
            config['consul']['id'] + '_check',
            check=check,
            service_id=config['consul']['id']
        )
        return True
    else:
        return False


class DownloadService(object):

    channel = None
    redis_client = None

    def supervise(self):
        if consul_declare(self.config):
            web_thread = threading.Thread(target=start_web, args=(self.config,))
            web_thread.start()

    def __init__(self, config_file=None, rabbitmq=True):
        self.logger = logging
        self.session = None
        self.bank = None
        self.download_callback = None
        with open(config_file, 'r') as ymlfile:
            self.config = yaml.load(ymlfile, Loader=yaml.FullLoader)
            Utils.service_config_override(self.config)

        Zipkin.set_config(self.config)

        if 'log_config' in self.config:
            for handler in list(self.config['log_config']['handlers'].keys()):
                self.config['log_config']['handlers'][handler] = dict(self.config['log_config']['handlers'][handler])
            logging.config.dictConfig(self.config['log_config'])
            self.logger = logging.getLogger('biomaj')

        if not self.redis_client:
            self.redis_client = redis.StrictRedis(host=self.config['redis']['host'],
                                                  port=self.config['redis']['port'],
                                                  db=self.config['redis']['db'],
                                                  decode_responses=True)

        if rabbitmq and not self.channel:
            connection = None
            rabbitmq_port = self.config['rabbitmq']['port']
            rabbitmq_user = self.config['rabbitmq']['user']
            rabbitmq_password = self.config['rabbitmq']['password']
            rabbitmq_vhost = self.config['rabbitmq']['virtual_host']
            if rabbitmq_user:
                credentials = pika.PlainCredentials(rabbitmq_user, rabbitmq_password)
                connection = pika.BlockingConnection(pika.ConnectionParameters(self.config['rabbitmq']['host'], rabbitmq_port, rabbitmq_vhost, credentials, heartbeat_interval=0))
            else:
                connection = pika.BlockingConnection(pika.ConnectionParameters(self.config['rabbitmq']['host'], heartbeat_interval=0))
            self.channel = connection.channel()
            self.logger.info('Download service started')

    def close(self):
        if self.channel:
            try:
                self.channel.close()
            except Exception as e:
                logging.warn('Download:Service:Exception:' + str(e))

    def on_download_callback(self, func):
        self.download_callback = func

    def get_handler(self, protocol_name, server, remote_dir, remote_files=[],
                    credentials=None, http_parse=None, http_method=None, param=None,
                    proxy=None, proxy_auth='',
                    save_as=None, timeout_download=None, offline_dir=None,
                    options={}):
        protocol = downmessage_pb2.DownloadFile.Protocol.Value(protocol_name.upper())
        downloader = None
        if protocol in [0, 1]:  # FTP, SFTP
            downloader = CurlDownload(protocol_name, server, remote_dir)
        if protocol in [2, 3]:  # HTTP, HTTPS (could be factored with previous case)
            downloader = CurlDownload(protocol_name, server, remote_dir, http_parse)
        if protocol == 4:  # DirectFTP
            downloader = DirectFTPDownload("ftp", server, '/')
        if protocol == 5:  # DirectHTTP
            downloader = DirectHTTPDownload("http", server, '/')
        if protocol == 6:  # DirectHTTPS
            downloader = DirectHTTPDownload("https", server, '/')
        if protocol == 10:  # DirectFTPS
            downloader = DirectFTPDownload('ftps', server, '/')
        if protocol == 7:  # Local
            downloader = LocalDownload(remote_dir)
        if protocol == 8:  # RSYNC
            downloader = RSYNCDownload(server, remote_dir)
        if protocol == 9:  # iRods
            downloader = IRODSDownload(server, remote_dir)
        if downloader is None:
            return None

        for remote_file in remote_files:
            if remote_file['save_as']:
                save_as = remote_file['save_as']
        # For direct protocols, we only keep base name
        if protocol in DIRECT_PROTOCOLS:
            tmp_remote = []
            for remote_file in remote_files:
                tmp_remote.append(remote_file['name'])
            remote_files = tmp_remote

        if http_method is not None:
            downloader.set_method(http_method)

        if offline_dir:
            downloader.set_offline_dir(offline_dir)

        if proxy is not None and proxy:
            downloader.set_proxy(proxy, proxy_auth)

        if timeout_download is not None and timeout_download:
            downloader.set_timeout(timeout_download)

        if credentials:
            downloader.set_credentials(credentials)

        if save_as:
            downloader.set_save_as(save_as)

        if param:
            downloader.set_param(param)

        downloader.set_server(server)

        # Set the name of the BioMAJ protocol to which we respond.
        downloader.set_protocol(protocol_name)

        if options is not None:
            self.logger.debug("Received options: " + str(options))
            downloader.set_options(options)

        downloader.logger = self.logger
        downloader.set_files_to_download(remote_files)
        return downloader

    def _get_handler(self, biomaj_file_info):
        """
        Get a protocol download handler
        """

        protocol = biomaj_file_info.remote_file.protocol
        server = biomaj_file_info.remote_file.server
        remote_dir = biomaj_file_info.remote_file.remote_dir

        protocol_name = downmessage_pb2.DownloadFile.Protocol.Name(protocol).lower()
        self.logger.debug('%s request to download from %s://%s' % (biomaj_file_info.bank, protocol_name, server))

        remote_files = []
        for remote_file in biomaj_file_info.remote_file.files:
            remote_files.append({
                                'name': remote_file.name,
                                'save_as': remote_file.save_as,
                                'year': remote_file.metadata.year,
                                'month': remote_file.metadata.month,
                                'day': remote_file.metadata.day,
                                'root': remote_file.root
                                })

        proxy = None
        proxy_auth = ''
        if biomaj_file_info.proxy is not None:
            proxy = biomaj_file_info.proxy.proxy
            proxy_auth = biomaj_file_info.proxy.proxy_auth

        params = None
        if biomaj_file_info.remote_file.param:
            params = {}
            for param in biomaj_file_info.remote_file.param:
                params[param.name] = param.value
        return self.get_handler(protocol_name, server, remote_dir,
                                remote_files=remote_files,
                                credentials=biomaj_file_info.remote_file.credentials,
                                http_parse=biomaj_file_info.remote_file.http_parse,
                                http_method=downmessage_pb2.DownloadFile.HTTP_METHOD.Name(biomaj_file_info.http_method),
                                param=params,
                                proxy=proxy,
                                proxy_auth=proxy_auth,
                                save_as=biomaj_file_info.remote_file.save_as,
                                timeout_download=biomaj_file_info.timeout_download,
                                offline_dir=biomaj_file_info.local_dir,
                                options=biomaj_file_info.options
                                )

    def clean(self, biomaj_file_info=None):
        '''
        Clean session and download info
        '''
        session = self.session
        bank = self.bank
        if biomaj_file_info:
            session = biomaj_file_info.session
            bank = biomaj_file_info.bank

        self.logger.debug('Clean %s session %s' % (bank, session))
        self.redis_client.delete(self.config['redis']['prefix'] + ':' + bank + ':session:' + session)
        self.redis_client.delete(self.config['redis']['prefix'] + ':' + bank + ':session:' + session + ':error')
        self.redis_client.delete(self.config['redis']['prefix'] + ':' + bank + ':session:' + session + ':progress')
        self.redis_client.delete(self.config['redis']['prefix'] + ':' + bank + ':session:' + session + ':files')
        self.redis_client.delete(self.config['redis']['prefix'] + ':' + bank + ':session:' + session + ':error:info')

    def _create_session(self, bank):
        '''
        Creates a unique session
        '''
        self.session = str(uuid.uuid4())
        self.redis_client.set(self.config['redis']['prefix'] + ':' + bank + ':session:' + self.session, 1)
        self.logger.debug('Create %s new session %s' % (bank, self.session))
        self.bank = bank
        return self.session

    def download_errors(self, biomaj_file_info):
        '''
        Get errors
        '''
        errors = []
        error = self.redis_client.rpop(self.config['redis']['prefix'] + ':' + biomaj_file_info.bank + ':session:' + biomaj_file_info.session + ':error:info')
        while error:
            errors.append(error)
            error = self.redis_client.rpop(self.config['redis']['prefix'] + ':' + biomaj_file_info.bank + ':session:' + biomaj_file_info.session + ':error:info')
        return errors

    def download_status(self, biomaj_file_info):
        '''
        Get current status
        '''

        error = self.redis_client.get(self.config['redis']['prefix'] + ':' + biomaj_file_info.bank + ':session:' + biomaj_file_info.session + ':error')
        progress = self.redis_client.get(self.config['redis']['prefix'] + ':' + biomaj_file_info.bank + ':session:' + biomaj_file_info.session + ':progress')
        if error is None:
            error = -1
        if progress is None:
            progress = -1
        return (int(progress), int(error))

    def list_status(self, biomaj_file_info):

        list_progress = self.redis_client.get(self.config['redis']['prefix'] + ':' + biomaj_file_info.bank + ':session:' + biomaj_file_info.session + ':progress')
        if list_progress:
            return True
        else:
            return False

    def list_result(self, biomaj_file_info, protobuf_decode=True):
        '''
        Get file list result
        '''

        file_list = self.redis_client.get(self.config['redis']['prefix'] + ':' + biomaj_file_info.bank + ':session:' + biomaj_file_info.session + ':files')
        if protobuf_decode:
            file_list_pb2 = downmessage_pb2.FileList()
            file_list_pb2.ParseFromString(file_list_pb2)
            return file_list_pb2

        return file_list

    def _list(self, download_handler, biomaj_file_info):
        '''
        List remote content, no session management
        '''
        file_list = []
        dir_list = []
        file_list_pb2 = downmessage_pb2.FileList()

        try:
            (file_list, dir_list) = download_handler.list()
            download_handler.match(biomaj_file_info.remote_file.matches, file_list, dir_list)
        except Exception as e:
            self.logger.error('List exception for bank %s: %s' % (biomaj_file_info.bank, str(e)))
            self.redis_client.set(self.config['redis']['prefix'] + ':' + biomaj_file_info.bank + ':session:' + biomaj_file_info.session + ':error', 1)
            self.redis_client.lpush(self.config['redis']['prefix'] + ':' + biomaj_file_info.bank + ':session:' + biomaj_file_info.session + ':error:info', str(e))
        else:
            self.logger.debug('End of download for %s session %s' % (biomaj_file_info.bank, biomaj_file_info.session))
            for file_elt in download_handler.files_to_download:
                file_pb2 = file_list_pb2.files.add()
                file_pb2.name = file_elt['name']
                file_pb2.root = file_elt['root']
                if 'save_as' in file_elt:
                    file_pb2.save_as = file_elt['save_as']
                if 'url' in file_elt:
                    file_pb2.url = file_elt['url']
                if 'param' in file_elt and file_elt['param']:
                    for key in list(file_elt['param'].keys()):
                        param = file_list_pb2.param.add()
                        param.name = key
                        param.value = file_elt['param'][key]
                metadata = downmessage_pb2.File.MetaData()
                metadata.permissions = file_elt['permissions']
                metadata.group = file_elt['group']
                metadata.size = int(file_elt['size'])
                metadata.hash = file_elt['hash']
                metadata.year = int(file_elt['year'])
                metadata.month = int(file_elt['month'])
                metadata.day = int(file_elt['day'])
                if 'format' in file_elt:
                    metadata.format = file_elt['format']
                file_pb2.metadata.MergeFrom(metadata)
        return file_list_pb2

    def list(self, biomaj_file_info):
        '''
        List remote content
        '''
        self.logger.debug('New list request %s session %s' % (biomaj_file_info.bank, biomaj_file_info.session))
        session = self.redis_client.get(self.config['redis']['prefix'] + ':' + biomaj_file_info.bank + ':session:' + biomaj_file_info.session)
        if not session:
            self.logger.debug('Session %s for bank %s has expired, skipping download of %s' % (biomaj_file_info.session, biomaj_file_info.bank, biomaj_file_info.remote_file.files))
            return
        download_handler = self._get_handler(biomaj_file_info)
        if download_handler is None:
            self.logger.error('Could not get a handler for %s with session %s' % (biomaj_file_info.bank, biomaj_file_info.session))

        file_list_pb2 = self._list(download_handler, biomaj_file_info)

        self.redis_client.set(self.config['redis']['prefix'] + ':' + biomaj_file_info.bank + ':session:' + biomaj_file_info.session + ':files', str(file_list_pb2.SerializeToString()))
        self.redis_client.incr(self.config['redis']['prefix'] + ':' + biomaj_file_info.bank + ':session:' + biomaj_file_info.session + ':progress')

    def local_download(self, biomaj_file_info):
        '''
        Download files, no session
        '''
        download_handler = self._get_handler(biomaj_file_info)
        if download_handler is None:
            self.logger.error('Could not get a handler for %s with session %s' % (biomaj_file_info.bank, biomaj_file_info.session))
        downloaded_files = download_handler.download(biomaj_file_info.local_dir)
        self.logger.debug("Downloaded " + str(len(downloaded_files)) + " file in " + biomaj_file_info.local_dir)
        self.get_file_info(biomaj_file_info.local_dir, downloaded_files)
        return downloaded_files

    def download(self, biomaj_file_info):
        '''
        Download files

        Store in redis the progress and count of errors under:
         - prefix:bank_name:session:session_id:error
         - prefix:bank_name:session:session_id:progress
        '''

        self.logger.debug('New download request %s session %s' % (biomaj_file_info.bank, biomaj_file_info.session))
        session = self.redis_client.get(self.config['redis']['prefix'] + ':' + biomaj_file_info.bank + ':session:' + biomaj_file_info.session)
        if not session:
            self.logger.debug('Session %s for bank %s has expired, skipping download of %s' % (biomaj_file_info.session, biomaj_file_info.bank, biomaj_file_info.remote_file.files))
            return
        downloaded_files = []
        try:
            downloaded_files = self.local_download(biomaj_file_info)
        except Exception as e:
            self.logger.exception("Download error:%s:%s:%s" % (biomaj_file_info.bank, biomaj_file_info.session, str(e)))
            session = self.redis_client.get(self.config['redis']['prefix'] + ':' + biomaj_file_info.bank + ':session:' + biomaj_file_info.session)
            if session:
                # If session deleted, do not track
                self.redis_client.incr(self.config['redis']['prefix'] + ':' + biomaj_file_info.bank + ':session:' + biomaj_file_info.session + ':error')
                self.redis_client.lpush(self.config['redis']['prefix'] + ':' + biomaj_file_info.bank + ':session:' + biomaj_file_info.session + ':error:info', str(e))
        else:
            self.logger.debug('End of download for %s session %s' % (biomaj_file_info.bank, biomaj_file_info.session))

        session = self.redis_client.get(self.config['redis']['prefix'] + ':' + biomaj_file_info.bank + ':session:' + biomaj_file_info.session)
        if session:
            # If session deleted, do not track
            self.redis_client.incr(self.config['redis']['prefix'] + ':' + biomaj_file_info.bank + ':session:' + biomaj_file_info.session + ':progress')
        return downloaded_files

    def get_file_info(self, local_dir, downloaded_files):
        if downloaded_files is None:
            return
        for downloaded_file in downloaded_files:
            # file_dir = local_dir + '/' + os.path.dirname(downloaded_file['save_as'])
            file_path = local_dir + '/' + downloaded_file['save_as']
            fstat = os.stat(file_path)
            downloaded_file['permissions'] = str(fstat.st_mode)
            downloaded_file['group'] = str(fstat.st_gid)
            downloaded_file['user'] = str(fstat.st_uid)
            downloaded_file['size'] = str(fstat.st_size)
            fstat_mtime = datetime.datetime.fromtimestamp(fstat.st_mtime)
            downloaded_file['month'] = fstat_mtime.month
            downloaded_file['day'] = fstat_mtime.day
            downloaded_file['year'] = fstat_mtime.year

    def ask_download(self, biomaj_info_file):
        self.channel.basic_publish(
            exchange='',
            routing_key='biomajdownload',
            body=biomaj_info_file.SerializeToString(),
            properties=pika.BasicProperties(
                # make message persistent
                delivery_mode=2
            ))

    def callback_messages(self, ch, method, properties, body):
        '''
        Manage download and send ACK message
        '''
        try:
            operation = downmessage_pb2.Operation()
            operation.ParseFromString(body)
            message = operation.download
            span = None
            if operation.trace and operation.trace.trace_id:
                url = str(message.remote_file.protocol) + ':' + str(message.remote_file.server) + ':' + str(message.remote_file.remote_dir)
                span = Zipkin('biomaj-download-executor', str(message.remote_file.server), trace_id=operation.trace.trace_id, parent_id=operation.trace.span_id)
                span.add_binary_annotation('url', url)
                span.add_binary_annotation('local_dir', str(message.local_dir))

            self.logger.debug('Received message: %s' % (message))
            if operation.type == 0:
                message = operation.download
                self.logger.debug('List operation %s, %s' % (message.bank, message.session))
                if len(message.remote_file.matches) == 0:
                    self.logger.error('No pattern match for a list operation')
                else:
                    self.list(message)
            elif operation.type == 1:
                message = operation.download
                self.logger.debug('Download operation %s, %s' % (message.bank, message.session))
                downloaded_files = self.download(message)
                if self.download_callback is not None:
                    self.download_callback(message.bank, downloaded_files)
            else:
                self.logger.warn('Wrong message type, skipping')
            if span:
                span.trace()
        except Exception as e:
            self.logger.error('Error with message: %s' % (str(e)))
            traceback.print_exc()
        ch.basic_ack(delivery_tag=method.delivery_tag)

    def wait_for_messages(self):
        '''
        Loop queue waiting for messages
        '''
        self.channel.queue_declare(queue='biomajdownload', durable=True)
        self.channel.basic_qos(prefetch_count=1)
        self.channel.basic_consume(
            self.callback_messages,
            queue='biomajdownload')
        self.channel.start_consuming()



'''
Message consumer for download requests
'''

import os
import logging

import requests
import yaml
import consul

from biomaj_download.downloadservice import DownloadService
from biomaj_core.utils import Utils

config_file = 'config.yml'
if 'BIOMAJ_CONFIG' in os.environ:
        config_file = os.environ['BIOMAJ_CONFIG']

config = None
with open(config_file, 'r') as ymlfile:
    config = yaml.load(ymlfile, Loader=yaml.FullLoader)
    Utils.service_config_override(config)


def on_download(bank, downloaded_files):
    metrics = []
    if 'prometheus' in config and not config['prometheus']:
        return
    if not downloaded_files:
        metric = {'bank': bank, 'error': 1}
        metrics.append(metrics)
    else:
        for downloaded_file in downloaded_files:
            metric = {'bank': bank}
            if 'error' in downloaded_file and downloaded_file['error']:
                metric['error'] = 1
            else:
                metric['size'] = downloaded_file['size']
                metric['download_time'] = downloaded_file['download_time']
            if 'hostname' in config['web']:
                metric['host'] = config['web']['hostname']
            metrics.append(metric)
        proxy = Utils.get_service_endpoint(config, 'download')
        r = requests.post(proxy + '/api/download/metrics', json = metrics)


download = DownloadService(config_file)
download.on_download_callback(on_download)
download.supervise()
download.wait_for_messages()



# from future import standard_library
# standard_library.install_aliases()
# from builtins import str
import re
import subprocess

from biomaj_download.download.interface import DownloadInterface


class RSYNCDownload(DownloadInterface):
    '''
    Base class to download files from rsync
    protocol = rsync
    server =
    remote.dir =

    remote.files =
    '''

    # This is used to forge the command
    real_protocol = "rsync"

    def __init__(self, server, rootdir):
        DownloadInterface.__init__(self)
        self.logger.debug('Download')
        # If rootdir is not given, we are in local mode. In this case, server
        # is interpreted as rootdir
        self.local_mode = not rootdir
        if not self.local_mode:
            self.server = server  # name of the remote server
            self.rootdir = rootdir  # directory on the remote server
        else:
            self.server = None
            self.rootdir = server

    def _append_file_to_download(self, rfile):
        if 'root' not in rfile or not rfile['root']:
            rfile['root'] = self.rootdir
        super(RSYNCDownload, self)._append_file_to_download(rfile)

    def _remote_file_name(self, rfile):
        # rfile['root'] is set to self.rootdir. We don't use os.path.join
        # because rfile['name'] may starts with /
        url = rfile['root'] + "/" + rfile['name']
        if not self.local_mode:
            url = self.server + ":" + url
        return re.sub("/{2,}", "/", url)

    def _network_configuration(self):
        '''
        Perform some configuration before network operations (list and
        download).
        '''
        pass

    def _download(self, file_path, rfile):
        error = False
        err_code = ''
        url = self._remote_file_name(rfile)
        # Create the rsync command
        if self.credentials:
            cmd = str(self.real_protocol) + " " + str(self.credentials) + "@" + url + " " + str(file_path)
        else:
            cmd = str(self.real_protocol) + " " + url + " " + str(file_path)
        self.logger.debug('RSYNC:RSYNC DOWNLOAD:' + cmd)
        # Launch the command (we are in offline_dir)
        try:
            p = subprocess.Popen(cmd, stdin=subprocess.PIPE, stderr=subprocess.PIPE, stdout=subprocess.PIPE, shell=True)
            stdout, stderr = p.communicate()
            err_code = p.returncode
            self.test_stderr_rsync_message(stderr)
            self.test_stderr_rsync_error(stderr)
        except ExceptionRsync as e:
            self.logger.error(str(self.real_protocol) + " error:" + str(e))
        if err_code != 0:
            self.logger.error('Error while downloading ' + rfile["name"] + ' - ' + str(err_code))
            error = True
        if error:
            return error

        # Our part is done so call parent _download
        return super(RSYNCDownload, self)._download(file_path, rfile)

    def test_stderr_rsync_error(self, stderr):
        stderr = str(stderr.decode('utf-8'))
        if "rsync error" in str(stderr):
            reason = stderr.split(str(self.real_protocol) + " error:")[1].split("\n")[0]
            raise ExceptionRsync(reason)

    def test_stderr_rsync_message(self, stderr):
        stderr = str(stderr.decode('utf-8'))
        if "rsync:" in str(stderr):
            reason = stderr.split(str(self.real_protocol) + ":")[1].split("\n")[0]
            raise ExceptionRsync(reason)

    def list(self, directory=''):
        '''
        List server directory

        :return: dict of file and dirs in current directory with details
        '''
        err_code = None
        rfiles = []
        rdirs = []
        self.logger.debug('RSYNC:List')
        if self.local_mode:
            remote = str(self.rootdir) + str(directory)
        else:
            remote = str(self.server) + ":" + str(self.rootdir) + str(directory)
        if self.credentials:
            remote = str(self.credentials) + "@" + remote
        cmd = str(self.real_protocol) + " --list-only --no-motd " + remote
        try:
            p = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            list_rsync, err = p.communicate()
            self.test_stderr_rsync_message(err)
            self.test_stderr_rsync_error(err)
            err_code = p.returncode
            if err_code != 0:
                msg = 'Error while listing ' + remote + ' - ' + str(err_code)
                self.logger.error(msg)
                raise Exception(msg)
        except ExceptionRsync as e:
            msg = 'Error while listing ' + remote + ' - ' + str(e)
            self.logger.error(msg)
            raise e
        list_rsync = str(list_rsync.decode('utf-8'))
        lines = list_rsync.rstrip().split("\n")
        for line in lines:
            rfile = {}
            # rsync LIST output is separated by \n
            parts = line.split()
            if not parts:
                continue
            date = parts[2].split('/')
            rfile['permissions'] = parts[0]
            rfile['size'] = int(parts[1].replace(',', ''))
            rfile['month'] = int(date[1])
            rfile['day'] = int(date[2])
            rfile['year'] = int(date[0])
            rfile['name'] = parts[4]
            is_dir = False
            if re.match('^d', rfile['permissions']):
                is_dir = True

            if not is_dir:
                rfiles.append(rfile)
            else:
                rdirs.append(rfile)

        return (rfiles, rdirs)


class ExceptionRsync(Exception):
    def __init__(self, exception_reason):
        self.exception_reason = exception_reason

    def __str__(self):
        return self.exception_reason



import os
import logging
import datetime
import time
import re
import copy

import tenacity
from simpleeval import simple_eval, ast

from biomaj_core.utils import Utils
from biomaj_core.config import BiomajConfig


class _FakeLock(object):
    '''
    Fake lock for downloaders not called by a Downloadthread
    '''

    def __init__(self):
        pass

    def acquire(self):
        pass

    def release(self):
        pass


class DownloadInterface(object):
    '''
    Main interface that all downloaders must extend.

    The methods are divided into 2 broad categories:
      - setters which act on properties of the downloader; those methods are
        important in microservice mode
      - file operations which are used to list and match remote files, download
        them, etc.

    Usually, it is enough to overload list, _append_file_to_download and
    _download.

    TODO:
      - the purpose of some setters (set_server, set_protocol) is not clear
        since a subclass cannot always change those parameters arbitrarily
      - chroot is not used in BioMaJ
    '''

    files_num_threads = 4

    #
    # Constants to parse retryer
    #
    # Note that due to the current implementation of operators, tenacity allows
    # nonsensical operations. For example the following snippets are valid:
    # stop_after_attempt(1, 2) + 4
    # stop_after_attempt(1, 2) + stop_none.
    # Of course, trying to use those wait policies will raise cryptic errors.
    # The situation is similar for stop conditions.
    # See https://github.com/jd/tenacity/issues/211.
    # To avoid such errors, we test the objects in _set_retryer.
    #
    # Another confusing issue is that stop_never is an object (instance of the
    # class _stop_never). For parsing, if we consider stop_never as a
    # function then both "stop_never" and "stop_never()" are parsed correctly
    # but the later raises error. Considering it has a name is slightly more
    # clear (since then we must write "stop_none" as we do when we use tenacity
    # directly). For consistency, we create a name for wait_none (as an
    # instance of the class wait_none).
    #

    # Functions available when parsing stop condition: those are constructors
    # of stop conditions classes (then using them will create objects). Note
    # that there is an exception for stop_never.
    ALL_STOP_CONDITIONS = {
        # "stop_never": tenacity.stop._stop_never,  # In case, we want to use it like a function (see above)
        "stop_when_event_set": tenacity.stop_when_event_set,
        "stop_after_attempt": tenacity.stop_after_attempt,
        "stop_after_delay": tenacity.stop_after_delay,
        "stop_any": tenacity.stop_any,  # Similar to |
        "stop_all": tenacity.stop_all,  # Similar to &
    }

    # tenacity.stop_never is an instance of _stop_never, not a class so we
    # import it as a name.
    ALL_STOP_NAMES = {
        "stop_never": tenacity.stop_never,
    }

    # Operators for stop conditions: | means to stop if one of the conditions
    # is True, & means to stop if all the conditions are True.
    ALL_STOP_OPERATORS = {
        ast.BitOr: tenacity.stop.stop_base.__or__,
        ast.BitAnd: tenacity.stop.stop_base.__and__,
    }

    # Functions available when parsing wait policy: those are constructors
    # of wait policies classes (then using them will create objects). Note
    # that there is an exception for wait_none.
    ALL_WAIT_POLICIES = {
        # "wait_none": tenacity.wait_none,  # In case, we want to use it like a function (see above)
        "wait_fixed": tenacity.wait_fixed,
        "wait_random": tenacity.wait_random,
        "wait_incrementing": tenacity.wait_incrementing,
        "wait_exponential": tenacity.wait_exponential,
        "wait_random_exponential": tenacity.wait_random_exponential,
        "wait_combine": tenacity.wait_combine,  # Sum of wait policies (similar to +)
        "wait_chain": tenacity.wait_chain,  # Give a list of wait policies (one for each attempt)
    }

    # Create an instance of wait_none to use it like a constant.
    ALL_WAIT_NAMES = {
        "wait_none": tenacity.wait.wait_none()
    }

    # Operators for wait policies: + means to sum waiting times of wait
    # policies.
    ALL_WAIT_OPERATORS = {
        ast.Add: tenacity.wait.wait_base.__add__
    }

    @staticmethod
    def is_true(download_error):
        """Method used by retryer to determine if we should retry to downlaod a
        file based on the return value of method:`_download` (passed as the
        argument): we must retry while this value is True.

        See method:`_set_retryer`.
        """
        return download_error is True

    @staticmethod
    def return_last_value(retry_state):
        """Method used by the retryer to determine the return value of the
        retryer: we return the result of the last attempt.

        See method:`_set_retryer`.
        """
        return retry_state.outcome.result()

    def __init__(self):
        # This variable defines the protocol as passed by the config file (i.e.
        # this is directftp for DirectFTPDownload). It is used by the workflow
        # to send the download message so it must be set.
        self.protocol = None
        self.config = None
        self.files_to_download = []
        self.files_to_copy = []
        self.error = False
        self.credentials = None
        # bank name
        self.bank = None
        self.mkdir_lock = _FakeLock()
        self.kill_received = False
        self.proxy = None
        # 24h timeout
        self.timeout = 3600
        # Optional save target for single file downloaders
        self.save_as = None
        self.logger = logging.getLogger('biomaj')
        self.param = None
        self.method = None
        self.server = None
        self.offline_dir = None
        # Options
        self.options = {}  # This field is used to forge the download message
        self.skip_check_uncompress = False
        # TODO: Don't store default values in BiomajConfig.DEFAULTS for
        # wait_policy and stop_condition
        # Construct default retryer (may be replaced in set_options)
        self._set_retryer(
            BiomajConfig.DEFAULTS["stop_condition"],
            BiomajConfig.DEFAULTS["wait_policy"]
        )

    #
    # Setters for downloader
    #

    def set_offline_dir(self, offline_dir):
        self.offline_dir = offline_dir

    def set_server(self, server):
        self.server = server

    def set_protocol(self, protocol):
        """
        Method used by DownloadService to set the protocol. This value is
        passed from the config file so is not always a real protocol (for
        instance it can be "directhttp" for a direct downloader).
        """
        self.protocol = protocol

    def set_param(self, param):
        self.param = param

    def set_timeout(self, timeout):
        if isinstance(timeout, int):
            self.timeout = timeout
        else:
            try:
                self.timeout = int(timeout)
            except Exception:
                logging.error('Timeout is not a valid integer, skipping')

    def set_save_as(self, save_as):
        self.save_as = save_as

    def set_proxy(self, proxy, proxy_auth=None):
        '''
        Use a proxy to connect to remote servers

        :param proxy: proxy to use (see http://curl.haxx.se/libcurl/c/CURLOPT_PROXY.html for format)
        :type proxy: str
        :param proxy_auth: proxy authentication if any (user:password)
        :type proxy_auth: str
        '''
        self.proxy = proxy
        self.proxy_auth = proxy_auth

    def set_method(self, method):
        self.method = method

    def set_credentials(self, userpwd):
        '''
        Set credentials in format user:pwd

        :param userpwd: credentials
        :type userpwd: str
        '''
        self.credentials = userpwd

    def set_options(self, options):
        """
        Set download options.

        Subclasses that override this method must call this implementation.
        """
        # Copy the option dict
        self.options = options
        if "skip_check_uncompress" in options:
            self.skip_check_uncompress = Utils.to_bool(options["skip_check_uncompress"])
        # If stop_condition or wait_policy is specified, we reconstruct the retryer
        if "stop_condition" or "wait_policy" in options:
            stop_condition = options.get("stop_condition", BiomajConfig.DEFAULTS["stop_condition"])
            wait_policy = options.get("wait_policy", BiomajConfig.DEFAULTS["wait_policy"])
            self._set_retryer(stop_condition, wait_policy)

    def _set_retryer(self, stop_condition, wait_policy):
        """
        Add a retryer to retry the current download if it fails.
        """
        # Try to construct stop condition
        if isinstance(stop_condition, tenacity.stop.stop_base):
            # Use the value directly
            stop_cond = stop_condition
        elif isinstance(stop_condition, str):
            # Try to parse the string
            try:
                stop_cond = simple_eval(stop_condition,
                                        functions=self.ALL_STOP_CONDITIONS,
                                        operators=self.ALL_STOP_OPERATORS,
                                        names=self.ALL_STOP_NAMES)
                # Check that it is an instance of stop_base
                if not isinstance(stop_cond, tenacity.stop.stop_base):
                    raise ValueError(stop_condition + " doesn't yield a stop condition")
                # Test that this is a correct stop condition by calling it.
                # We use a deepcopy to be sure to not alter the object (even
                # if it seems that calling a wait policy doesn't modify it).
                try:
                    s = copy.deepcopy(stop_cond)
                    s(tenacity.compat.make_retry_state(0, 0))
                except Exception:
                    raise ValueError(stop_condition + " doesn't yield a stop condition")
            except Exception as e:
                raise ValueError("Error while parsing stop condition: %s" % e)
        else:
            raise TypeError("Expected tenacity.stop.stop_base or string, got %s" % type(stop_condition))
        # Try to construct wait policy
        if isinstance(wait_policy, tenacity.wait.wait_base):
            # Use the value directly
            wait_pol = wait_policy
        elif isinstance(wait_policy, str):
            # Try to parse the string
            try:
                wait_pol = simple_eval(wait_policy,
                                       functions=self.ALL_WAIT_POLICIES,
                                       operators=self.ALL_WAIT_OPERATORS,
                                       names=self.ALL_WAIT_NAMES)
                # Check that it is an instance of wait_base
                if not isinstance(wait_pol, tenacity.wait.wait_base):
                    raise ValueError(wait_policy + " doesn't yield a wait policy")
                # Test that this is a correct wait policy by calling it.
                # We use a deepcopy to be sure to not alter the object (even
                # if it seems that calling a stop condition doesn't modify it).
                try:
                    w = copy.deepcopy(wait_pol)
                    w(tenacity.compat.make_retry_state(0, 0))
                except Exception:
                    raise ValueError(wait_policy + " doesn't yield a wait policy")
            except Exception as e:
                raise ValueError("Error while parsing wait policy: %s" % e)
        else:
            raise TypeError("Expected tenacity.stop.wait_base or string, got %s" % type(wait_policy))

        self.retryer = tenacity.Retrying(
            stop=stop_cond,
            wait=wait_pol,
            retry_error_callback=self.return_last_value,
            retry=tenacity.retry_if_result(self.is_true),
            reraise=True
        )

    #
    # File operations (match, list, download) and associated hook methods
    #

    def _append_file_to_download(self, rfile):
        """
        Add a file to the download list and check its properties (this method
        is called in `match` and `set_files_to_download`).

        Downloaders can override this to add some properties to the file (for
        instance, most of them will add "root").
        """
        # Add properties to the file if needed (for safety)
        if 'save_as' not in rfile or rfile['save_as'] is None:
            rfile['save_as'] = rfile['name']
        if self.param:
            if 'param' not in rfile or not rfile['param']:
                rfile['param'] = self.param
        # Remove duplicate */* if any
        rfile['name'] = re.sub('//+', '/', rfile['name'])
        self.files_to_download.append(rfile)

    def set_files_to_download(self, files):
        """
        Convenience method to set the list of files to download.
        """
        self.files_to_download = []
        for file_to_download in files:
            self._append_file_to_download(file_to_download)

    def match(self, patterns, file_list, dir_list=None, prefix='', submatch=False):
        '''
        Find files matching patterns. Sets instance variable files_to_download.

        :param patterns: regexps to match
        :type patterns: list
        :param file_list: list of files to match
        :type file_list: list
        :param dir_list: sub directories in current dir
        :type dir_list: list
        :param prefix: directory prefix
        :type prefix: str
        :param submatch: first call to match, or called from match
        :type submatch: bool
        '''
        self.logger.debug('Download:File:RegExp:' + str(patterns))
        if dir_list is None:
            dir_list = []

        if not submatch:
            self.files_to_download = []
        for pattern in patterns:
            subdirs_pattern = pattern.split('/')
            if len(subdirs_pattern) > 1:
                # Pattern contains sub directories
                subdir = subdirs_pattern[0]
                if subdir == '^':
                    subdirs_pattern = subdirs_pattern[1:]
                    subdir = subdirs_pattern[0]
                # If getting all, get all files
                if pattern == '**/*':
                    for rfile in file_list:
                        if prefix != '':
                            rfile['name'] = prefix + '/' + rfile['name']
                        self._append_file_to_download(rfile)
                        self.logger.debug('Download:File:MatchRegExp:' + rfile['name'])
                    return
                for direlt in dir_list:
                    subdir = direlt['name']
                    self.logger.debug('Download:File:Subdir:Check:' + subdir)
                    if pattern == '**/*':
                        (subfile_list, subdirs_list) = self.list(prefix + '/' + subdir + '/')
                        self.match([pattern], subfile_list, subdirs_list, prefix + '/' + subdir, True)
                        for rfile in file_list:
                            if pattern == '**/*' or re.match(pattern, rfile['name']):
                                if prefix != '':
                                    rfile['name'] = prefix + '/' + rfile['name']
                                self._append_file_to_download(rfile)
                                self.logger.debug('Download:File:MatchRegExp:' + rfile['name'])
                    else:
                        if re.match(subdirs_pattern[0], subdir):
                            self.logger.debug('Download:File:Subdir:Match:' + subdir)
                            # subdir match the beginning of the pattern
                            # check match in subdir
                            (subfile_list, subdirs_list) = self.list(prefix + '/' + subdir + '/')
                            self.match(['/'.join(subdirs_pattern[1:])], subfile_list, subdirs_list, prefix + '/' + subdir, True)

            else:
                for rfile in file_list:
                    if re.match(pattern, rfile['name']):
                        if prefix != '':
                            rfile['name'] = prefix + '/' + rfile['name']
                        self._append_file_to_download(rfile)
                        self.logger.debug('Download:File:MatchRegExp:' + rfile['name'])

        if not submatch and len(self.files_to_download) == 0:
            raise Exception('no file found matching expressions')

    def set_permissions(self, file_path, file_info):
        '''
        Sets file attributes to remote ones
        '''
        if file_info['year'] and file_info['month'] and file_info['day']:
            ftime = datetime.date(file_info['year'], file_info['month'], file_info['day'])
            settime = time.mktime(ftime.timetuple())
            os.utime(file_path, (settime, settime))

    def download_or_copy(self, available_files, root_dir, check_exists=True):
        '''
        If a file to download is available in available_files, copy it instead of downloading it.

        Update the instance variables files_to_download and files_to_copy

        :param available_files: list of files available in root_dir
        :type available files: list
        :param root_dir: directory where files are available
        :type root_dir: str
        :param check_exists: checks if file exists locally
        :type check_exists: bool
        '''

        self.files_to_copy = []
        # In such case, it forces the download again
        if not available_files:
            return
        available_files.sort(key=lambda x: x['name'])
        self.files_to_download.sort(key=lambda x: x['name'])

        new_files_to_download = []

        test1_tuples = set((d['name'], d['year'], d['month'], d['day'], d['size']) for d in self.files_to_download)
        test2_tuples = set((d['name'], d['year'], d['month'], d['day'], d['size']) for d in available_files)
        new_or_modified_files = [t for t in test1_tuples if t not in test2_tuples]
        new_or_modified_files.sort(key=lambda x: x[0])
        index = 0

        if len(new_or_modified_files) > 0:
            self.logger.debug('Number of remote files: %d' % (len(self.files_to_download)))
            self.logger.debug('Number of local files: %d' % (len(available_files)))
            self.logger.debug('Number of files new or modified: %d' % (len(new_or_modified_files)))
            for dfile in self.files_to_download:
                if index < len(new_or_modified_files) and \
                        dfile['name'] == new_or_modified_files[index][0]:
                    new_files_to_download.append(dfile)
                    index += 1
                else:
                    fileName = dfile["name"]
                    if dfile["name"].startswith('/'):
                        fileName = dfile["name"][1:]
                    if not check_exists or os.path.exists(os.path.join(root_dir, fileName)):
                        dfile['root'] = root_dir
                        self.logger.debug('Copy file instead of downloading it: %s' % (os.path.join(root_dir, dfile['name'])))
                        self.files_to_copy.append(dfile)
                    else:
                        new_files_to_download.append(dfile)
        else:
            # Copy everything
            for dfile in self.files_to_download:
                fileName = dfile["name"]
                if dfile["name"].startswith('/'):
                    fileName = dfile["name"][1:]
                if not check_exists or os.path.exists(os.path.join(root_dir, fileName)):
                    dfile['root'] = root_dir
                    self.files_to_copy.append(dfile)
                else:
                    new_files_to_download.append(dfile)

        self.set_files_to_download(new_files_to_download)

    def _download(self, file_path, rfile):
        '''
        Download one file and return False in case of success and True
        otherwise.

        Subclasses that override this method must call this implementation
        at the end to perform test on archives.

        Note that this method is executed inside a retryer.
        '''
        error = False
        # Check that the archive is correct
        if not self.skip_check_uncompress:
            archive_status = Utils.archive_check(file_path)
            if not archive_status:
                self.logger.error('Archive is invalid or corrupted, deleting file and retrying download')
                error = True
                if os.path.exists(file_path):
                    os.remove(file_path)
        return error

    def _network_configuration(self):
        '''
        Perform some configuration before network operations (list and
        download). This must be implemented in subclasses.
        '''
        raise NotImplementedError()

    def download(self, local_dir, keep_dirs=True):
        '''
        Download remote files to local_dir

        :param local_dir: Directory where files should be downloaded
        :type local_dir: str
        :param keep_dirs: keep file name directory structure or copy file in local_dir directly
        :param keep_dirs: bool
        :return: list of downloaded files
        '''
        self.logger.debug(self.__class__.__name__ + ':Download')
        self._network_configuration()
        nb_files = len(self.files_to_download)
        cur_files = 1
        self.offline_dir = local_dir
        for rfile in self.files_to_download:
            if self.kill_received:
                raise Exception('Kill request received, exiting')
            # Determine where to store file (directory and name)
            file_dir = local_dir
            if keep_dirs:
                file_dir = local_dir + '/' + os.path.dirname(rfile['save_as'])
            if file_dir[-1] == "/":
                file_path = file_dir + os.path.basename(rfile['save_as'])
            else:
                file_path = file_dir + '/' + os.path.basename(rfile['save_as'])

            # For unit tests only, workflow will take in charge directory
            # creation before to avoid thread multi access
            if not os.path.exists(file_dir):
                os.makedirs(file_dir)

            msg = self.__class__.__name__ + ':Download:Progress:'
            msg += str(cur_files) + '/' + str(nb_files)
            msg += ' downloading file ' + rfile['name'] + ' save as ' + rfile['save_as']
            self.logger.debug(msg)
            cur_files += 1
            start_time = datetime.datetime.now()
            start_time = time.mktime(start_time.timetuple())
            error = self.retryer(self._download, file_path, rfile)
            if error:
                rfile['download_time'] = 0
                rfile['error'] = True
                raise Exception(self.__class__.__name__ + ":Download:Error:" + rfile["name"])
            else:
                end_time = datetime.datetime.now()
                end_time = time.mktime(end_time.timetuple())
                rfile['download_time'] = end_time - start_time
            # Set permissions
            self.set_permissions(file_path, rfile)

        return self.files_to_download

    def list(self):
        '''
        List directory

        :return: tuple of file list and dir list
        '''
        pass

    def chroot(self, cwd):
        '''
        Change directory
        '''
        pass

    def close(self):
        '''
        Close connection
        '''
        pass



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: downmessage.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='downmessage.proto',
  package='biomaj.download',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n\x11\x64ownmessage.proto\x12\x0f\x62iomaj.download\"\x9d\x02\n\x04\x46ile\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\x0c\n\x04root\x18\x02 \x01(\t\x12\x0f\n\x07save_as\x18\x03 \x01(\t\x12\x0b\n\x03url\x18\x04 \x01(\t\x12\x30\n\x08metadata\x18\x05 \x01(\x0b\x32\x1e.biomaj.download.File.MetaData\x1a\xa8\x01\n\x08MetaData\x12\x13\n\x0bpermissions\x18\x01 \x01(\t\x12\r\n\x05group\x18\x02 \x01(\t\x12\x0c\n\x04size\x18\x03 \x01(\x03\x12\x0c\n\x04hash\x18\x04 \x01(\t\x12\x0c\n\x04year\x18\x05 \x01(\x05\x12\r\n\x05month\x18\x06 \x01(\x05\x12\x0b\n\x03\x64\x61y\x18\x07 \x01(\x05\x12\x0e\n\x06\x66ormat\x18\x08 \x01(\t\x12\x0b\n\x03md5\x18\t \x01(\t\x12\x15\n\rdownload_time\x18\n \x01(\x03\"0\n\x08\x46ileList\x12$\n\x05\x66iles\x18\x01 \x03(\x0b\x32\x15.biomaj.download.File\"\xaa\x02\n\tOperation\x12\x32\n\x04type\x18\x01 \x02(\x0e\x32$.biomaj.download.Operation.OPERATION\x12/\n\x08\x64ownload\x18\x02 \x01(\x0b\x32\x1d.biomaj.download.DownloadFile\x12)\n\x07process\x18\x03 \x01(\x0b\x32\x18.biomaj.download.Process\x12/\n\x05trace\x18\x04 \x01(\x0b\x32 .biomaj.download.Operation.Trace\x1a*\n\x05Trace\x12\x10\n\x08trace_id\x18\x01 \x02(\t\x12\x0f\n\x07span_id\x18\x02 \x02(\t\"0\n\tOPERATION\x12\x08\n\x04LIST\x10\x00\x12\x0c\n\x08\x44OWNLOAD\x10\x01\x12\x0b\n\x07PROCESS\x10\x02\"\x17\n\x07Process\x12\x0c\n\x04\x65xec\x18\x01 \x02(\t\"\x94\x0b\n\x0c\x44ownloadFile\x12\x0c\n\x04\x62\x61nk\x18\x01 \x02(\t\x12\x0f\n\x07session\x18\x02 \x02(\t\x12\x11\n\tlocal_dir\x18\x03 \x02(\t\x12\x18\n\x10timeout_download\x18\x04 \x01(\x05\x12=\n\x0bremote_file\x18\x05 \x02(\x0b\x32(.biomaj.download.DownloadFile.RemoteFile\x12\x32\n\x05proxy\x18\x06 \x01(\x0b\x32#.biomaj.download.DownloadFile.Proxy\x12\x43\n\x0bhttp_method\x18\x08 \x01(\x0e\x32).biomaj.download.DownloadFile.HTTP_METHOD:\x03GET\x12;\n\x07options\x18\t \x03(\x0b\x32*.biomaj.download.DownloadFile.OptionsEntry\x1a$\n\x05Param\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\r\n\x05value\x18\x02 \x02(\t\x1a\xcd\x03\n\tHttpParse\x12\x91\x01\n\x08\x64ir_line\x18\x01 \x02(\t:\x7f<img[\\s]+src=\"[\\S]+\"[\\s]+alt=\"\\[DIR\\]\"[\\s]*/?>[\\s]*<a[\\s]+href=\"([\\S]+)/\"[\\s]*>.*([\\d]{2}-[\\w\\d]{2,5}-[\\d]{4}\\s[\\d]{2}:[\\d]{2})\x12\xa5\x01\n\tfile_line\x18\x02 \x02(\t:\x91\x01<img[\\s]+src=\"[\\S]+\"[\\s]+alt=\"\\[[\\s]+\\]\"[\\s]*/?>[\\s]<a[\\s]+href=\"([\\S]+)\".*([\\d]{2}-[\\w\\d]{2,5}-[\\d]{4}\\s[\\d]{2}:[\\d]{2})[\\s]+([\\d\\.]+[MKG]{0,1})\x12\x13\n\x08\x64ir_name\x18\x03 \x02(\x05:\x01\x31\x12\x13\n\x08\x64ir_date\x18\x04 \x02(\x05:\x01\x32\x12\x14\n\tfile_name\x18\x05 \x02(\x05:\x01\x31\x12\x14\n\tfile_date\x18\x06 \x02(\x05:\x01\x32\x12\x18\n\x10\x66ile_date_format\x18\x07 \x01(\t\x12\x14\n\tfile_size\x18\x08 \x02(\x05:\x01\x33\x1a\xb8\x02\n\nRemoteFile\x12$\n\x05\x66iles\x18\x01 \x03(\x0b\x32\x15.biomaj.download.File\x12\x38\n\x08protocol\x18\x02 \x02(\x0e\x32&.biomaj.download.DownloadFile.Protocol\x12\x0e\n\x06server\x18\x03 \x02(\t\x12\x12\n\nremote_dir\x18\x04 \x02(\t\x12\x0f\n\x07save_as\x18\x05 \x01(\t\x12\x32\n\x05param\x18\x06 \x03(\x0b\x32#.biomaj.download.DownloadFile.Param\x12;\n\nhttp_parse\x18\x07 \x01(\x0b\x32\'.biomaj.download.DownloadFile.HttpParse\x12\x13\n\x0b\x63redentials\x18\x08 \x01(\t\x12\x0f\n\x07matches\x18\t \x03(\t\x1a*\n\x05Proxy\x12\r\n\x05proxy\x18\x01 \x02(\t\x12\x12\n\nproxy_auth\x18\x02 \x01(\t\x1a.\n\x0cOptionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x93\x01\n\x08Protocol\x12\x07\n\x03\x46TP\x10\x00\x12\x08\n\x04\x46TPS\x10\x01\x12\x08\n\x04HTTP\x10\x02\x12\t\n\x05HTTPS\x10\x03\x12\r\n\tDIRECTFTP\x10\x04\x12\x0e\n\nDIRECTHTTP\x10\x05\x12\x0f\n\x0b\x44IRECTHTTPS\x10\x06\x12\t\n\x05LOCAL\x10\x07\x12\t\n\x05RSYNC\x10\x08\x12\t\n\x05IRODS\x10\t\x12\x0e\n\nDIRECTFTPS\x10\n\" \n\x0bHTTP_METHOD\x12\x07\n\x03GET\x10\x00\x12\x08\n\x04POST\x10\x01')
)



_OPERATION_OPERATION = _descriptor.EnumDescriptor(
  name='OPERATION',
  full_name='biomaj.download.Operation.OPERATION',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='LIST', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DOWNLOAD', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PROCESS', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=627,
  serialized_end=675,
)
_sym_db.RegisterEnumDescriptor(_OPERATION_OPERATION)

_DOWNLOADFILE_PROTOCOL = _descriptor.EnumDescriptor(
  name='Protocol',
  full_name='biomaj.download.DownloadFile.Protocol',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='FTP', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FTPS', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HTTP', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HTTPS', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DIRECTFTP', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DIRECTHTTP', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DIRECTHTTPS', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOCAL', index=7, number=7,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RSYNC', index=8, number=8,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IRODS', index=9, number=9,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DIRECTFTPS', index=10, number=10,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1950,
  serialized_end=2097,
)
_sym_db.RegisterEnumDescriptor(_DOWNLOADFILE_PROTOCOL)

_DOWNLOADFILE_HTTP_METHOD = _descriptor.EnumDescriptor(
  name='HTTP_METHOD',
  full_name='biomaj.download.DownloadFile.HTTP_METHOD',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='GET', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='POST', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2099,
  serialized_end=2131,
)
_sym_db.RegisterEnumDescriptor(_DOWNLOADFILE_HTTP_METHOD)


_FILE_METADATA = _descriptor.Descriptor(
  name='MetaData',
  full_name='biomaj.download.File.MetaData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='permissions', full_name='biomaj.download.File.MetaData.permissions', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='group', full_name='biomaj.download.File.MetaData.group', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='size', full_name='biomaj.download.File.MetaData.size', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hash', full_name='biomaj.download.File.MetaData.hash', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='year', full_name='biomaj.download.File.MetaData.year', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='month', full_name='biomaj.download.File.MetaData.month', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='day', full_name='biomaj.download.File.MetaData.day', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='format', full_name='biomaj.download.File.MetaData.format', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='md5', full_name='biomaj.download.File.MetaData.md5', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='download_time', full_name='biomaj.download.File.MetaData.download_time', index=9,
      number=10, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=156,
  serialized_end=324,
)

_FILE = _descriptor.Descriptor(
  name='File',
  full_name='biomaj.download.File',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='biomaj.download.File.name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='root', full_name='biomaj.download.File.root', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='save_as', full_name='biomaj.download.File.save_as', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='url', full_name='biomaj.download.File.url', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='metadata', full_name='biomaj.download.File.metadata', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_FILE_METADATA, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=39,
  serialized_end=324,
)


_FILELIST = _descriptor.Descriptor(
  name='FileList',
  full_name='biomaj.download.FileList',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='files', full_name='biomaj.download.FileList.files', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=326,
  serialized_end=374,
)


_OPERATION_TRACE = _descriptor.Descriptor(
  name='Trace',
  full_name='biomaj.download.Operation.Trace',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='trace_id', full_name='biomaj.download.Operation.Trace.trace_id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='span_id', full_name='biomaj.download.Operation.Trace.span_id', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=583,
  serialized_end=625,
)

_OPERATION = _descriptor.Descriptor(
  name='Operation',
  full_name='biomaj.download.Operation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='biomaj.download.Operation.type', index=0,
      number=1, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='download', full_name='biomaj.download.Operation.download', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='process', full_name='biomaj.download.Operation.process', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='trace', full_name='biomaj.download.Operation.trace', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_OPERATION_TRACE, ],
  enum_types=[
    _OPERATION_OPERATION,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=377,
  serialized_end=675,
)


_PROCESS = _descriptor.Descriptor(
  name='Process',
  full_name='biomaj.download.Process',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='exec', full_name='biomaj.download.Process.exec', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=677,
  serialized_end=700,
)


_DOWNLOADFILE_PARAM = _descriptor.Descriptor(
  name='Param',
  full_name='biomaj.download.DownloadFile.Param',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='biomaj.download.DownloadFile.Param.name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='biomaj.download.DownloadFile.Param.value', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1040,
  serialized_end=1076,
)

_DOWNLOADFILE_HTTPPARSE = _descriptor.Descriptor(
  name='HttpParse',
  full_name='biomaj.download.DownloadFile.HttpParse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='dir_line', full_name='biomaj.download.DownloadFile.HttpParse.dir_line', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=True, default_value=_b("<img[\\s]+src=\"[\\S]+\"[\\s]+alt=\"\\[DIR\\]\"[\\s]*/?>[\\s]*<a[\\s]+href=\"([\\S]+)/\"[\\s]*>.*([\\d]{2}-[\\w\\d]{2,5}-[\\d]{4}\\s[\\d]{2}:[\\d]{2})").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file_line', full_name='biomaj.download.DownloadFile.HttpParse.file_line', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=True, default_value=_b("<img[\\s]+src=\"[\\S]+\"[\\s]+alt=\"\\[[\\s]+\\]\"[\\s]*/?>[\\s]<a[\\s]+href=\"([\\S]+)\".*([\\d]{2}-[\\w\\d]{2,5}-[\\d]{4}\\s[\\d]{2}:[\\d]{2})[\\s]+([\\d\\.]+[MKG]{0,1})").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dir_name', full_name='biomaj.download.DownloadFile.HttpParse.dir_name', index=2,
      number=3, type=5, cpp_type=1, label=2,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dir_date', full_name='biomaj.download.DownloadFile.HttpParse.dir_date', index=3,
      number=4, type=5, cpp_type=1, label=2,
      has_default_value=True, default_value=2,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file_name', full_name='biomaj.download.DownloadFile.HttpParse.file_name', index=4,
      number=5, type=5, cpp_type=1, label=2,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file_date', full_name='biomaj.download.DownloadFile.HttpParse.file_date', index=5,
      number=6, type=5, cpp_type=1, label=2,
      has_default_value=True, default_value=2,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file_date_format', full_name='biomaj.download.DownloadFile.HttpParse.file_date_format', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file_size', full_name='biomaj.download.DownloadFile.HttpParse.file_size', index=7,
      number=8, type=5, cpp_type=1, label=2,
      has_default_value=True, default_value=3,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1079,
  serialized_end=1540,
)

_DOWNLOADFILE_REMOTEFILE = _descriptor.Descriptor(
  name='RemoteFile',
  full_name='biomaj.download.DownloadFile.RemoteFile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='files', full_name='biomaj.download.DownloadFile.RemoteFile.files', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='protocol', full_name='biomaj.download.DownloadFile.RemoteFile.protocol', index=1,
      number=2, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='server', full_name='biomaj.download.DownloadFile.RemoteFile.server', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='remote_dir', full_name='biomaj.download.DownloadFile.RemoteFile.remote_dir', index=3,
      number=4, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='save_as', full_name='biomaj.download.DownloadFile.RemoteFile.save_as', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='param', full_name='biomaj.download.DownloadFile.RemoteFile.param', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='http_parse', full_name='biomaj.download.DownloadFile.RemoteFile.http_parse', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='credentials', full_name='biomaj.download.DownloadFile.RemoteFile.credentials', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='matches', full_name='biomaj.download.DownloadFile.RemoteFile.matches', index=8,
      number=9, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1543,
  serialized_end=1855,
)

_DOWNLOADFILE_PROXY = _descriptor.Descriptor(
  name='Proxy',
  full_name='biomaj.download.DownloadFile.Proxy',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='proxy', full_name='biomaj.download.DownloadFile.Proxy.proxy', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proxy_auth', full_name='biomaj.download.DownloadFile.Proxy.proxy_auth', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1857,
  serialized_end=1899,
)

_DOWNLOADFILE_OPTIONSENTRY = _descriptor.Descriptor(
  name='OptionsEntry',
  full_name='biomaj.download.DownloadFile.OptionsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='biomaj.download.DownloadFile.OptionsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='biomaj.download.DownloadFile.OptionsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1901,
  serialized_end=1947,
)

_DOWNLOADFILE = _descriptor.Descriptor(
  name='DownloadFile',
  full_name='biomaj.download.DownloadFile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='bank', full_name='biomaj.download.DownloadFile.bank', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='session', full_name='biomaj.download.DownloadFile.session', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='local_dir', full_name='biomaj.download.DownloadFile.local_dir', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timeout_download', full_name='biomaj.download.DownloadFile.timeout_download', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='remote_file', full_name='biomaj.download.DownloadFile.remote_file', index=4,
      number=5, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proxy', full_name='biomaj.download.DownloadFile.proxy', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='http_method', full_name='biomaj.download.DownloadFile.http_method', index=6,
      number=8, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='options', full_name='biomaj.download.DownloadFile.options', index=7,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_DOWNLOADFILE_PARAM, _DOWNLOADFILE_HTTPPARSE, _DOWNLOADFILE_REMOTEFILE, _DOWNLOADFILE_PROXY, _DOWNLOADFILE_OPTIONSENTRY, ],
  enum_types=[
    _DOWNLOADFILE_PROTOCOL,
    _DOWNLOADFILE_HTTP_METHOD,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=703,
  serialized_end=2131,
)

_FILE_METADATA.containing_type = _FILE
_FILE.fields_by_name['metadata'].message_type = _FILE_METADATA
_FILELIST.fields_by_name['files'].message_type = _FILE
_OPERATION_TRACE.containing_type = _OPERATION
_OPERATION.fields_by_name['type'].enum_type = _OPERATION_OPERATION
_OPERATION.fields_by_name['download'].message_type = _DOWNLOADFILE
_OPERATION.fields_by_name['process'].message_type = _PROCESS
_OPERATION.fields_by_name['trace'].message_type = _OPERATION_TRACE
_OPERATION_OPERATION.containing_type = _OPERATION
_DOWNLOADFILE_PARAM.containing_type = _DOWNLOADFILE
_DOWNLOADFILE_HTTPPARSE.containing_type = _DOWNLOADFILE
_DOWNLOADFILE_REMOTEFILE.fields_by_name['files'].message_type = _FILE
_DOWNLOADFILE_REMOTEFILE.fields_by_name['protocol'].enum_type = _DOWNLOADFILE_PROTOCOL
_DOWNLOADFILE_REMOTEFILE.fields_by_name['param'].message_type = _DOWNLOADFILE_PARAM
_DOWNLOADFILE_REMOTEFILE.fields_by_name['http_parse'].message_type = _DOWNLOADFILE_HTTPPARSE
_DOWNLOADFILE_REMOTEFILE.containing_type = _DOWNLOADFILE
_DOWNLOADFILE_PROXY.containing_type = _DOWNLOADFILE
_DOWNLOADFILE_OPTIONSENTRY.containing_type = _DOWNLOADFILE
_DOWNLOADFILE.fields_by_name['remote_file'].message_type = _DOWNLOADFILE_REMOTEFILE
_DOWNLOADFILE.fields_by_name['proxy'].message_type = _DOWNLOADFILE_PROXY
_DOWNLOADFILE.fields_by_name['http_method'].enum_type = _DOWNLOADFILE_HTTP_METHOD
_DOWNLOADFILE.fields_by_name['options'].message_type = _DOWNLOADFILE_OPTIONSENTRY
_DOWNLOADFILE_PROTOCOL.containing_type = _DOWNLOADFILE
_DOWNLOADFILE_HTTP_METHOD.containing_type = _DOWNLOADFILE
DESCRIPTOR.message_types_by_name['File'] = _FILE
DESCRIPTOR.message_types_by_name['FileList'] = _FILELIST
DESCRIPTOR.message_types_by_name['Operation'] = _OPERATION
DESCRIPTOR.message_types_by_name['Process'] = _PROCESS
DESCRIPTOR.message_types_by_name['DownloadFile'] = _DOWNLOADFILE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

File = _reflection.GeneratedProtocolMessageType('File', (_message.Message,), dict(

  MetaData = _reflection.GeneratedProtocolMessageType('MetaData', (_message.Message,), dict(
    DESCRIPTOR = _FILE_METADATA,
    __module__ = 'downmessage_pb2'
    # @@protoc_insertion_point(class_scope:biomaj.download.File.MetaData)
    ))
  ,
  DESCRIPTOR = _FILE,
  __module__ = 'downmessage_pb2'
  # @@protoc_insertion_point(class_scope:biomaj.download.File)
  ))
_sym_db.RegisterMessage(File)
_sym_db.RegisterMessage(File.MetaData)

FileList = _reflection.GeneratedProtocolMessageType('FileList', (_message.Message,), dict(
  DESCRIPTOR = _FILELIST,
  __module__ = 'downmessage_pb2'
  # @@protoc_insertion_point(class_scope:biomaj.download.FileList)
  ))
_sym_db.RegisterMessage(FileList)

Operation = _reflection.GeneratedProtocolMessageType('Operation', (_message.Message,), dict(

  Trace = _reflection.GeneratedProtocolMessageType('Trace', (_message.Message,), dict(
    DESCRIPTOR = _OPERATION_TRACE,
    __module__ = 'downmessage_pb2'
    # @@protoc_insertion_point(class_scope:biomaj.download.Operation.Trace)
    ))
  ,
  DESCRIPTOR = _OPERATION,
  __module__ = 'downmessage_pb2'
  # @@protoc_insertion_point(class_scope:biomaj.download.Operation)
  ))
_sym_db.RegisterMessage(Operation)
_sym_db.RegisterMessage(Operation.Trace)

Process = _reflection.GeneratedProtocolMessageType('Process', (_message.Message,), dict(
  DESCRIPTOR = _PROCESS,
  __module__ = 'downmessage_pb2'
  # @@protoc_insertion_point(class_scope:biomaj.download.Process)
  ))
_sym_db.RegisterMessage(Process)

DownloadFile = _reflection.GeneratedProtocolMessageType('DownloadFile', (_message.Message,), dict(

  Param = _reflection.GeneratedProtocolMessageType('Param', (_message.Message,), dict(
    DESCRIPTOR = _DOWNLOADFILE_PARAM,
    __module__ = 'downmessage_pb2'
    # @@protoc_insertion_point(class_scope:biomaj.download.DownloadFile.Param)
    ))
  ,

  HttpParse = _reflection.GeneratedProtocolMessageType('HttpParse', (_message.Message,), dict(
    DESCRIPTOR = _DOWNLOADFILE_HTTPPARSE,
    __module__ = 'downmessage_pb2'
    # @@protoc_insertion_point(class_scope:biomaj.download.DownloadFile.HttpParse)
    ))
  ,

  RemoteFile = _reflection.GeneratedProtocolMessageType('RemoteFile', (_message.Message,), dict(
    DESCRIPTOR = _DOWNLOADFILE_REMOTEFILE,
    __module__ = 'downmessage_pb2'
    # @@protoc_insertion_point(class_scope:biomaj.download.DownloadFile.RemoteFile)
    ))
  ,

  Proxy = _reflection.GeneratedProtocolMessageType('Proxy', (_message.Message,), dict(
    DESCRIPTOR = _DOWNLOADFILE_PROXY,
    __module__ = 'downmessage_pb2'
    # @@protoc_insertion_point(class_scope:biomaj.download.DownloadFile.Proxy)
    ))
  ,

  OptionsEntry = _reflection.GeneratedProtocolMessageType('OptionsEntry', (_message.Message,), dict(
    DESCRIPTOR = _DOWNLOADFILE_OPTIONSENTRY,
    __module__ = 'downmessage_pb2'
    # @@protoc_insertion_point(class_scope:biomaj.download.DownloadFile.OptionsEntry)
    ))
  ,
  DESCRIPTOR = _DOWNLOADFILE,
  __module__ = 'downmessage_pb2'
  # @@protoc_insertion_point(class_scope:biomaj.download.DownloadFile)
  ))
_sym_db.RegisterMessage(DownloadFile)
_sym_db.RegisterMessage(DownloadFile.Param)
_sym_db.RegisterMessage(DownloadFile.HttpParse)
_sym_db.RegisterMessage(DownloadFile.RemoteFile)
_sym_db.RegisterMessage(DownloadFile.Proxy)
_sym_db.RegisterMessage(DownloadFile.OptionsEntry)


_DOWNLOADFILE_OPTIONSENTRY._options = None
# @@protoc_insertion_point(module_scope)



import re
from datetime import datetime
import hashlib
import time
import stat
from urllib.parse import urlencode
from io import BytesIO

import pycurl
import ftputil

import humanfriendly

from biomaj_core.utils import Utils
from biomaj_core.config import BiomajConfig

from biomaj_download.download.interface import DownloadInterface


class HTTPParse(object):

    def __init__(self, dir_line, file_line, dir_name=1, dir_date=2, file_name=1, file_date=2, file_date_format=None, file_size=3):
        r'''
        http.parse.dir.line: <img[\s]+src="[\S]+"[\s]+alt="\[DIR\]"[\s]*/?>[\s]*<a[\s]+href="([\S]+)/"[\s]*>.*([\d]{2}-[\w\d]{2,5}-[\d]{4}\s[\d]{2}:[\d]{2})
        http.parse.file.line: <img[\s]+src="[\S]+"[\s]+alt="\[[\s]+\]"[\s]*/?>[\s]<a[\s]+href="([\S]+)".*([\d]{2}-[\w\d]{2,5}-[\d]{4}\s[\d]{2}:[\d]{2})[\s]+([\d\.]+[MKG]{0,1})
        http.group.dir.name: 1
        http.group.dir.date: 2
        http.group.file.name: 1
        http.group.file.date: 2
        http.group.file.size: 3
        '''
        self.dir_line = dir_line
        self.file_line = file_line
        self.dir_name = dir_name
        self.dir_date = dir_date
        self.file_name = file_name
        self.file_date = file_date
        self.file_size = file_size
        self.file_date_format = file_date_format


class CurlDownload(DownloadInterface):
    '''
    Base class to download files from FTP(S), HTTP(S) and SFTP.

    protocol=ftp
    server=ftp.ncbi.nih.gov
    remote.dir=/blast/db/FASTA/

    remote.files=^alu.*\\.gz$

    '''

    FTP_PROTOCOL_FAMILY = ["ftp", "ftps"]
    HTTP_PROTOCOL_FAMILY = ["http", "https"]
    SFTP_PROTOCOL_FAMILY = ["sftp"]
    ALL_PROTOCOLS = FTP_PROTOCOL_FAMILY + HTTP_PROTOCOL_FAMILY + SFTP_PROTOCOL_FAMILY

    # Utilities to parse ftp listings: UnixParser is the more common hence we
    # put it first
    ftp_listing_parsers = [
        ftputil.stat.UnixParser(),
        ftputil.stat.MSParser(),
    ]

    # Valid values for ftp_method options as string and int
    VALID_FTP_FILEMETHOD = {
        "default": pycurl.FTPMETHOD_DEFAULT,
        "multicwd": pycurl.FTPMETHOD_MULTICWD,
        "nocwd": pycurl.FTPMETHOD_NOCWD,
        "singlecwd": pycurl.FTPMETHOD_SINGLECWD,
    }

    # Valid values for ssh_new_host options as string and int
    VALID_SSH_NEW_HOST = {
        "reject": pycurl.KHSTAT_REJECT,
        "accept": pycurl.KHSTAT_FINE,
        "add": pycurl.KHSTAT_FINE_ADD_TO_FILE,
    }

    def __init__(self, curl_protocol, host, rootdir, http_parse=None):
        """
        Initialize a CurlDownloader.

        :param curl_protocol: (real) protocol to use
        :type curl_protocol: str (see :py:var:~CurlDownload.ALL_PROTOCOLS)

        :param host: server name
        :type host: str

        :param rootdir: base directory
        :type rootdir: str

        :param http_parse: object used to extract file information from HTML pages
        :type http_parse: py:class:HTTPParse.
        """
        DownloadInterface.__init__(self)
        self.logger.debug('Download')

        # Check for ssh support
        curl_opts_info = pycurl.version_info()
        curl_opts = []
        for opt in curl_opts_info:
            if isinstance(opt, tuple):
                for o in opt:
                    curl_opts.append(o)
            else:
                curl_opts.append(opt)
        if 'sftp' not in curl_opts:
            CurlDownload.ALL_PROTOCOLS = CurlDownload.FTP_PROTOCOL_FAMILY + CurlDownload.HTTP_PROTOCOL_FAMILY
            self.logger.warning("sftp not supported by curl: %s" % str(curl_opts_info))
        # Initialize curl_protocol.
        # Note that we don't change that field in set_protocol since this
        # method uses the protocol from the configuration file. It's not clear
        # what to do in this case.
        curl_protocol = curl_protocol.lower()
        if curl_protocol not in self.ALL_PROTOCOLS:
            raise ValueError("curl_protocol must be one of %s (case insensitive). Got %s." % (self.ALL_PROTOCOLS, curl_protocol))
        self.curl_protocol = curl_protocol
        # Initialize protocol specific constants
        if self.curl_protocol in self.FTP_PROTOCOL_FAMILY:
            self.protocol_family = "ftp"
            self._parse_result = self._ftp_parse_result
            self.ERRCODE_OK = [221, 226]
        elif self.curl_protocol in self.HTTP_PROTOCOL_FAMILY:
            self.protocol_family = "http"
            self._parse_result = self._http_parse_result
            self.ERRCODE_OK = [200]
        elif self.curl_protocol in self.SFTP_PROTOCOL_FAMILY:
            self.protocol_family = "sftp"
            self._parse_result = self._ftp_parse_result
            self.ERRCODE_OK = [0]
        else:  # Should not happen since we check before
            raise ValueError("Unknown protocol")
        self.rootdir = rootdir
        self.set_server(host)
        self.headers = {}
        self.http_parse = http_parse
        # Create the cURL object
        # This object is shared by all operations to use the cache.
        # Before using it, call method:`_basic_curl_configuration`.
        self.crl = pycurl.Curl()
        #
        # Initialize options
        #
        # Should we skip SSL verification (cURL -k/--insecure option)
        self.ssl_verifyhost = True
        self.ssl_verifypeer = True
        # Path to the certificate of the server (cURL --cacert option; PEM format)
        self.ssl_server_cert = None
        # Keep alive
        self.tcp_keepalive = 0
        # FTP method (cURL --ftp-method option)
        self.ftp_method = pycurl.FTPMETHOD_DEFAULT  # Use cURL default
        # TODO: Don't store default values in BiomajConfig.DEFAULTS for
        # ssh_hosts_file and ssh_new_hosts
        # known_hosts file
        self.ssh_hosts_file = BiomajConfig.DEFAULTS["ssh_hosts_file"]
        # How to treat unknown host
        self.ssh_new_host = self.VALID_SSH_NEW_HOST[BiomajConfig.DEFAULTS["ssh_new_host"]]
        # Allow redirections
        self.allow_redirections = True

    def _accept_new_hosts(self, known_key, found_key, match):
        # Key found in file: we can accept it
        # Don't use KHSTAT_FINE_ADD_TO_FILE because the key would be duplicated
        # See https://github.com/curl/curl/issues/4953.
        if match == pycurl.KHMATCH_OK:
            return pycurl.KHSTAT_FINE
        # Key not found in file: use the ssh_new_host option
        elif match == pycurl.KHMATCH_MISSING:
            return self.ssh_new_host
        # Key missmatch: the best option is to reject it
        else:
            return pycurl.KHSTAT_REJECT

    def _network_configuration(self):
        """
        Perform basic configuration (i.e. that doesn't depend on the
        operation: _download or list). This method should be called before any
        operation.
        """
        # Reset cURL options before setting them
        self.crl.reset()

        if self.proxy is not None:
            self.crl.setopt(pycurl.PROXY, self.proxy)
            if self.proxy_auth is not None:
                self.crl.setopt(pycurl.PROXYUSERPWD, self.proxy_auth)

        if self.credentials is not None:
            self.crl.setopt(pycurl.USERPWD, self.credentials)

        # Hosts file & function to decide for new hosts
        if self.curl_protocol in self.SFTP_PROTOCOL_FAMILY:
            self.crl.setopt(pycurl.SSH_KNOWNHOSTS, self.ssh_hosts_file)
            self.crl.setopt(pycurl.SSH_KEYFUNCTION, self._accept_new_hosts)

        # Configure TCP keepalive
        if self.tcp_keepalive:
            try:
                self.crl.setopt(pycurl.TCP_KEEPALIVE, True)
                self.crl.setopt(pycurl.TCP_KEEPIDLE, self.tcp_keepalive * 2)
                self.crl.setopt(pycurl.TCP_KEEPINTVL, self.tcp_keepalive)
            except Exception as e:
                self.logger.exception("TCP keepalive option failed: " + str(e))

        # Configure SSL verification (on some platforms, disabling
        # SSL_VERIFYPEER implies disabling SSL_VERIFYHOST so we set
        # SSL_VERIFYPEER after)
        self.crl.setopt(pycurl.SSL_VERIFYHOST, 2 if self.ssl_verifyhost else 0)
        self.crl.setopt(pycurl.SSL_VERIFYPEER, 1 if self.ssl_verifypeer else 0)
        if self.ssl_server_cert:
            # cacert is the name of the option for the curl command. The
            # corresponding cURL option is CURLOPT_CAINFO.
            # See https://curl.haxx.se/libcurl/c/CURLOPT_CAINFO.html
            # This is inspired by that https://curl.haxx.se/docs/sslcerts.html
            # (section "Certificate Verification", option 2) but the option
            # CURLOPT_CAPATH is for a directory of certificates.
            self.crl.setopt(pycurl.CAINFO, self.ssl_server_cert)

        # Configure ftp method
        self.crl.setopt(pycurl.FTP_FILEMETHOD, self.ftp_method)

        # Configure redirections
        self.crl.setopt(pycurl.FOLLOWLOCATION, self.allow_redirections)

        # Configure timeouts
        self.crl.setopt(pycurl.CONNECTTIMEOUT, 300)
        self.crl.setopt(pycurl.TIMEOUT, self.timeout)
        self.crl.setopt(pycurl.NOSIGNAL, 1)

        # Header function
        self.crl.setopt(pycurl.HEADERFUNCTION, self._header_function)

    def _header_function(self, header_line):
        # HTTP standard specifies that headers are encoded in iso-8859-1.
        # On Python 2, decoding step can be skipped.
        # On Python 3, decoding step is required.
        header_line = header_line.decode('iso-8859-1')

        # Header lines include the first status line (HTTP/1.x ...).
        # We are going to ignore all lines that don't have a colon in them.
        # This will botch headers that are split on multiple lines...
        if ':' not in header_line:
            return

        # Break the header line into header name and value.
        name, value = header_line.split(':', 1)

        # Remove whitespace that may be present.
        # Header lines include the trailing newline, and there may be whitespace
        # around the colon.
        name = name.strip()
        value = value.strip()

        # Header names are case insensitive.
        # Lowercase name here.
        name = name.lower()

        # Now we can actually record the header name and value.
        self.headers[name] = value

    def set_server(self, server):
        super(CurlDownload, self).set_server(server)
        self.url = self.curl_protocol + '://' + self.server

    def set_options(self, options):
        super(CurlDownload, self).set_options(options)
        if "ssl_verifyhost" in options:
            self.ssl_verifyhost = Utils.to_bool(options["ssl_verifyhost"])
        if "ssl_verifypeer" in options:
            self.ssl_verifypeer = Utils.to_bool(options["ssl_verifypeer"])
        if "ssl_server_cert" in options:
            self.ssl_server_cert = options["ssl_server_cert"]
        if "tcp_keepalive" in options:
            self.tcp_keepalive = Utils.to_int(options["tcp_keepalive"])
        if "ftp_method" in options:
            # raw_val is a string which contains the name of the option as in the CLI.
            # We always convert raw_val to a valid integer
            raw_val = options["ftp_method"].lower()
            if raw_val not in self.VALID_FTP_FILEMETHOD:
                raise ValueError("Invalid value for ftp_method")
            self.ftp_method = self.VALID_FTP_FILEMETHOD[raw_val]
        if "ssh_hosts_file" in options:
            self.ssh_hosts_file = options["ssh_hosts_file"]
        if "ssh_new_host" in options:
            raw_val = options["ssh_new_host"].lower()
            if raw_val not in self.VALID_SSH_NEW_HOST:
                raise ValueError("Invalid value for ssh_new_host")
            self.ssh_new_host = self.VALID_SSH_NEW_HOST[raw_val]
        if "allow_redirections" in options:
            self.allow_redirections = Utils.to_bool(options["allow_redirections"])

    def _append_file_to_download(self, rfile):
        # Add url and root to the file if needed (for safety)
        if 'url' not in rfile or not rfile['url']:
            rfile['url'] = self.url
        if 'root' not in rfile or not rfile['root']:
            rfile['root'] = self.rootdir
        super(CurlDownload, self)._append_file_to_download(rfile)

    def _file_url(self, rfile):
        # rfile['root'] is set to self.rootdir if needed but may be different.
        # We don't use os.path.join because rfile['name'] may starts with /
        url = self.url + '/' + rfile['root'] + rfile['name']
        url_elts = url.split('://')
        if len(url_elts) == 2:
            url_elts[1] = re.sub("/{2,}", "/", url_elts[1])
            return '://'.join(url_elts)
        return re.sub("/{2,}", "/", url)

    def _download(self, file_path, rfile):
        """
        Download one file and return False in case of success and True
        otherwise.

        This method is designed to work for FTP(S), HTTP(S) and SFTP.
        """
        error = True
        # Forge URL of remote file
        file_url = self._file_url(rfile)

        try:
            self.crl.setopt(pycurl.URL, file_url)
        except Exception:
            self.crl.setopt(pycurl.URL, file_url.encode('ascii', 'ignore'))

        # Create file and assign it to the pycurl object
        fp = open(file_path, "wb")
        self.crl.setopt(pycurl.WRITEFUNCTION, fp.write)

        # This is specific to HTTP
        if self.method == 'POST':
            # Form data must be provided already urlencoded.
            postfields = urlencode(self.param)
            # Sets request method to POST,
            # Content-Type header to application/x-www-form-urlencoded
            # and data to send in request body.
            self.crl.setopt(pycurl.POSTFIELDS, postfields)

        # Try download (we don't raise errors here since its the return value
        # ('error') that matters for the calling method; this is set to True
        # only in case of success).
        try:
            self.crl.perform()
            errcode = self.crl.getinfo(pycurl.RESPONSE_CODE)
            if int(errcode) not in self.ERRCODE_OK:
                error = True
                self.logger.error('Error while downloading ' + file_url + ' - ' + str(errcode))
            else:
                error = False
        except Exception as e:
            self.logger.error('Error while downloading ' + file_url + ' - ' + str(e))

        # Check if we were redirected
        if self.curl_protocol in self.HTTP_PROTOCOL_FAMILY:
            n_redirect = self.crl.getinfo(pycurl.REDIRECT_COUNT)
            if n_redirect:
                real_url = self.crl.getinfo(pycurl.EFFECTIVE_URL)
                redirect_time = self.crl.getinfo(pycurl.REDIRECT_TIME)
                msg_fmt = 'Download was redirected to %s (%i redirection(s), took %ss)'
                msg = msg_fmt % (real_url, n_redirect, redirect_time)
                self.logger.info(msg)

        # Close file
        fp.close()

        if error:
            return error

        # Our part is done so call parent _download
        return super(CurlDownload, self)._download(file_path, rfile)

    def list(self, directory=''):
        '''
        List remote directory

        :return: tuple of file and dirs in current directory with details

        This is a generic method for HTTP and FTP. The protocol-specific parts
        are done in _<protocol>_parse_result.
        '''
        dirbase = re.sub('//+', '/', self.rootdir + directory)
        dir_url = self.url + dirbase
        self.logger.debug('Download:List:' + dir_url)

        self._network_configuration()

        try:
            self.crl.setopt(pycurl.URL, dir_url)
        except Exception:
            self.crl.setopt(pycurl.URL, dir_url.encode('ascii', 'ignore'))

        # Create buffer and assign it to the pycurl object
        output = BytesIO()
        self.crl.setopt(pycurl.WRITEFUNCTION, output.write)

        # Try to list
        try:
            self.crl.perform()
            errcode = self.crl.getinfo(pycurl.RESPONSE_CODE)
            if int(errcode) not in self.ERRCODE_OK:
                msg = 'Error while listing ' + dir_url + ' - ' + str(errcode)
                self.logger.error(msg)
                raise Exception(msg)
        except Exception as e:
            msg = 'Error while listing ' + dir_url + ' - ' + str(e)
            self.logger.error(msg)
            raise e

        # Check if we were redirected
        if self.curl_protocol in self.HTTP_PROTOCOL_FAMILY:
            n_redirect = self.crl.getinfo(pycurl.REDIRECT_COUNT)
            if n_redirect:
                real_url = self.crl.getinfo(pycurl.EFFECTIVE_URL)
                redirect_time = self.crl.getinfo(pycurl.REDIRECT_TIME)
                msg_fmt = 'Download was redirected to %s (%i redirection(s), took %ss)'
                msg = msg_fmt % (real_url, n_redirect, redirect_time)
                self.logger.info(msg)

        # Figure out what encoding was sent with the response, if any.
        # Check against lowercased header name.
        encoding = None
        if 'content-type' in self.headers:
            content_type = self.headers['content-type'].lower()
            match = re.search(r'charset=(\S+)', content_type)
            if match:
                encoding = match.group(1)
        if encoding is None:
            # Default encoding for HTML is iso-8859-1.
            # Other content types may have different default encoding,
            # or in case of binary data, may have no encoding at all.
            encoding = 'iso-8859-1'

        # lets get the output in a string
        result = output.getvalue().decode(encoding)

        return self._parse_result(result)

    def _ftp_parse_result(self, result):
        # FTP LIST output is separated by \r\n
        # lets split the output in lines
        lines = re.split(r'[\n\r]+', result)
        # lets walk through each line
        rfiles = []
        rdirs = []

        for line in lines:
            # Skip empty lines (usually the last)
            if not line:
                continue
            # Parse the line
            for i, parser in enumerate(self.ftp_listing_parsers, 1):
                try:
                    stats = parser.parse_line(line)
                    break
                except ftputil.error.ParserError:
                    # If it's the last parser, re-raise the exception
                    if i == len(self.ftp_listing_parsers):
                        raise
                    else:
                        continue
            # Put stats in a dict
            rfile = {}
            rfile['name'] = stats._st_name
            # Reparse mode to a string
            rfile['permissions'] = stat.filemode(stats.st_mode)
            rfile['group'] = stats.st_gid
            rfile['user'] = stats.st_uid
            rfile['size'] = stats.st_size
            mtime = time.localtime(stats.st_mtime)
            rfile['year'] = mtime.tm_year
            rfile['month'] = mtime.tm_mon
            rfile['day'] = mtime.tm_mday
            rfile['hash'] = hashlib.md5(line.encode('utf-8')).hexdigest()

            is_link = stat.S_ISLNK(stats.st_mode)
            is_dir = stat.S_ISDIR(stats.st_mode)
            # Append links to dirs and files since we don't know what the
            # target is
            if is_link:
                rfiles.append(rfile)
                rdirs.append(rfile)
            else:
                if not is_dir:
                    rfiles.append(rfile)
                else:
                    rdirs.append(rfile)
        return (rfiles, rdirs)

    def _http_parse_result(self, result):
        rfiles = []
        rdirs = []

        dirs = re.findall(self.http_parse.dir_line, result)
        if dirs is not None and len(dirs) > 0:
            for founddir in dirs:
                rfile = {}
                rfile['permissions'] = ''
                rfile['group'] = ''
                rfile['user'] = ''
                rfile['size'] = 0
                date = founddir[self.http_parse.dir_date - 1]
                dirdate = date.split()
                parts = dirdate[0].split('-')
                # 19-Jul-2014 13:02
                rfile['month'] = Utils.month_to_num(parts[1])
                rfile['day'] = int(parts[0])
                rfile['year'] = int(parts[2])
                rfile['name'] = founddir[self.http_parse.dir_name - 1]
                rdirs.append(rfile)

        files = re.findall(self.http_parse.file_line, result)

        if files is not None and len(files) > 0:
            for foundfile in files:
                rfile = {}
                rfile['permissions'] = ''
                rfile['group'] = ''
                rfile['user'] = ''
                if self.http_parse.file_size != -1:
                    rfile['size'] = humanfriendly.parse_size(foundfile[self.http_parse.file_size - 1])
                else:
                    rfile['size'] = 0
                if self.http_parse.file_date != -1:
                    date = foundfile[self.http_parse.file_date - 1]
                    if self.http_parse.file_date_format:
                        date_object = datetime.strptime(date, self.http_parse.file_date_format.replace('%%', '%'))
                        rfile['month'] = date_object.month
                        rfile['day'] = date_object.day
                        rfile['year'] = date_object.year
                    else:
                        dirdate = date.split()
                        parts = dirdate[0].split('-')
                        # 19-Jul-2014 13:02
                        rfile['month'] = Utils.month_to_num(parts[1])
                        rfile['day'] = int(parts[0])
                        rfile['year'] = int(parts[2])
                else:
                    today = datetime.now()
                    date = '%s-%s-%s' % (today.year, today.month, today.day)
                    rfile['month'] = today.month
                    rfile['day'] = today.day
                    rfile['year'] = today.year
                rfile['name'] = foundfile[self.http_parse.file_name - 1]
                filehash = (rfile['name'] + str(date) + str(rfile['size'])).encode('utf-8')
                rfile['hash'] = hashlib.md5(filehash).hexdigest()
                rfiles.append(rfile)
        return (rfiles, rdirs)

    def close(self):
        if self.crl is not None:
            self.crl.close()
            self.crl = None



from biomaj_download.download.interface import DownloadInterface
from irods.session import iRODSSession
from irods.exception import iRODSException
from irods.models import DataObject, User


class IRODSDownload(DownloadInterface):

    # This is used only for messages
    real_protocol = "irods"

    def __init__(self, server, remote_dir):
        DownloadInterface.__init__(self)
        self.port = 1247
        self.remote_dir = remote_dir  # directory on the remote server including zone
        self.rootdir = remote_dir
        self.user = None
        self.password = None
        self.server = server
        self.zone = remote_dir.split("/")[0]

    def _append_file_to_download(self, rfile):
        if 'root' not in rfile or not rfile['root']:
            rfile['root'] = self.rootdir
        super(IRODSDownload, self)._append_file_to_download(rfile)

    def set_param(self, param):
        # param is a dictionary which has the following form :
        # {'password': u'biomaj', 'user': u'biomaj', 'port': u'port'}
        # port is optional
        self.param = param
        self.user = str(param['user'])
        self.password = str(param['password'])
        if 'port' in param:
            self.port = int(param['port'])

    def list(self, directory=''):
        self._network_configuration()
        rfiles = []
        rdirs = []
        rfile = {}
        date = None
        # Note that iRODS raise errors when trying to use the results
        # and not after query(). Therefore, the whole loop is inside
        # try/catch.
        try:
            query = self.session.query(DataObject.name, DataObject.size,
                                       DataObject.owner_name, DataObject.modify_time)
            results = query.filter(User.name == self.user).get_results()
            for result in results:
                # Avoid duplication
                if rfile != {} and rfile['name'] == str(result[DataObject.name]) \
                   and date == str(result[DataObject.modify_time]).split(" ")[0].split('-'):
                    continue
                rfile = {}
                date = str(result[DataObject.modify_time]).split(" ")[0].split('-')
                rfile['permissions'] = "-rwxr-xr-x"
                rfile['size'] = int(result[DataObject.size])
                rfile['month'] = int(date[1])
                rfile['day'] = int(date[2])
                rfile['year'] = int(date[0])
                rfile['name'] = str(result[DataObject.name])
                rfiles.append(rfile)
        except Exception as e:
            msg = 'Error while listing ' + self.remote_dir + ' - ' + repr(e)
            self.logger.error(msg)
            raise e
        finally:
            self.session.cleanup()
        return (rfiles, rdirs)

    def _network_configuration(self):
        self.session = iRODSSession(host=self.server, port=self.port,
                                    user=self.user, password=self.password,
                                    zone=self.zone)

    def _download(self, file_path, rfile):
        error = False
        self.logger.debug('IRODS:IRODS DOWNLOAD')
        try:
            # iRODS don't like multiple "/"
            if rfile['root'][-1] == "/":
                file_to_get = rfile['root'] + rfile['name']
            else:
                file_to_get = rfile['root'] + "/" + rfile['name']
            # Write the file to download in the wanted file_dir with the
            # python-irods iget
            self.session.data_objects.get(file_to_get, file_path)
        except iRODSException as e:
            error = True
            self.logger.error(self.__class__.__name__ + ":Download:Error:Can't get irods object " + file_to_get)
            self.logger.error(self.__class__.__name__ + ":Download:Error:" + repr(e))

        if error:
            return error

        # Our part is done so call parent _download
        return super(IRODSDownload, self)._download(file_path, rfile)
